@@ -1,26 +1,36 @@
 import asyncio
 import collections
 import warnings
-from typing import List  # noqa
-from typing import Awaitable, Callable, Generic, Optional, Tuple, TypeVar
+from typing import (
+    Awaitable,
+    Callable,
+    Deque,
+    Final,
+    Generic,
+    List,
+    Optional,
+    Tuple,
+    TypeVar,
+)
 
 from .base_protocol import BaseProtocol
-from .helpers import BaseTimerContext, set_exception, set_result
+from .helpers import (
+    _EXC_SENTINEL,
+    BaseTimerContext,
+    TimerNoop,
+    set_exception,
+    set_result,
+)
 from .log import internal_logger
 
-try:  # pragma: no cover
-    from typing import Deque  # noqa
-except ImportError:
-    from typing_extensions import Deque  # noqa
-
-
 __all__ = (
-    'EMPTY_PAYLOAD', 'EofStream', 'StreamReader', 'DataQueue',
-    'FlowControlDataQueue')
+    "EMPTY_PAYLOAD",
+    "EofStream",
+    "StreamReader",
+    "DataQueue",
+)
 
-DEFAULT_LIMIT = 2 ** 16
-
-_T = TypeVar('_T')
+_T = TypeVar("_T")
 
 
 class EofStream(Exception):
@@ -29,65 +39,63 @@ class EofStream(Exception):
 
 class AsyncStreamIterator(Generic[_T]):
 
+    __slots__ = ("read_func",)
+
     def __init__(self, read_func: Callable[[], Awaitable[_T]]) -> None:
         self.read_func = read_func
 
-    def __aiter__(self) -> 'AsyncStreamIterator[_T]':
+    def __aiter__(self) -> "AsyncStreamIterator[_T]":
         return self
 
     async def __anext__(self) -> _T:
         try:
             rv = await self.read_func()
         except EofStream:
-            raise StopAsyncIteration  # NOQA
-        if rv == b'':
-            raise StopAsyncIteration  # NOQA
+            raise StopAsyncIteration
+        if rv == b"":
+            raise StopAsyncIteration
         return rv
 
 
 class ChunkTupleAsyncStreamIterator:
 
-    def __init__(self, stream: 'StreamReader') -> None:
+    __slots__ = ("_stream",)
+
+    def __init__(self, stream: "StreamReader") -> None:
         self._stream = stream
 
-    def __aiter__(self) -> 'ChunkTupleAsyncStreamIterator':
+    def __aiter__(self) -> "ChunkTupleAsyncStreamIterator":
         return self
 
     async def __anext__(self) -> Tuple[bytes, bool]:
         rv = await self._stream.readchunk()
-        if rv == (b'', False):
-            raise StopAsyncIteration  # NOQA
+        if rv == (b"", False):
+            raise StopAsyncIteration
         return rv
 
 
 class AsyncStreamReaderMixin:
 
+    __slots__ = ()
+
     def __aiter__(self) -> AsyncStreamIterator[bytes]:
-        return AsyncStreamIterator(self.readline)  # type: ignore
+        return AsyncStreamIterator(self.readline)  # type: ignore[attr-defined]
 
     def iter_chunked(self, n: int) -> AsyncStreamIterator[bytes]:
-        """Returns an asynchronous iterator that yields chunks of size n.
-
-        Python-3.5 available for Python 3.5+ only
-        """
-        return AsyncStreamIterator(lambda: self.read(n))  # type: ignore
+        """Returns an asynchronous iterator that yields chunks of size n."""
+        return AsyncStreamIterator(lambda: self.read(n))  # type: ignore[attr-defined]
 
     def iter_any(self) -> AsyncStreamIterator[bytes]:
-        """Returns an asynchronous iterator that yields all the available
-        data as soon as it is received
-
-        Python-3.5 available for Python 3.5+ only
-        """
-        return AsyncStreamIterator(self.readany)  # type: ignore
+        """Yield all available data as soon as it is received."""
+        return AsyncStreamIterator(self.readany)  # type: ignore[attr-defined]
 
     def iter_chunks(self) -> ChunkTupleAsyncStreamIterator:
-        """Returns an asynchronous iterator that yields chunks of data
-        as they are received by the server. The yielded objects are tuples
-        of (bytes, bool) as returned by the StreamReader.readchunk method.
+        """Yield chunks of data as they are received by the server.
 
-        Python-3.5 available for Python 3.5+ only
+        The yielded objects are tuples
+        of (bytes, bool) as returned by the StreamReader.readchunk method.
         """
-        return ChunkTupleAsyncStreamIterator(self)  # type: ignore
+        return ChunkTupleAsyncStreamIterator(self)  # type: ignore[arg-type]
 
 
 class StreamReader(AsyncStreamReaderMixin):
@@ -104,12 +112,35 @@ class StreamReader(AsyncStreamReaderMixin):
 
     """
 
-    total_bytes = 0
+    __slots__ = (
+        "_protocol",
+        "_low_water",
+        "_high_water",
+        "_loop",
+        "_size",
+        "_cursor",
+        "_http_chunk_splits",
+        "_buffer",
+        "_buffer_offset",
+        "_eof",
+        "_waiter",
+        "_eof_waiter",
+        "_exception",
+        "_timer",
+        "_eof_callbacks",
+        "_eof_counter",
+        "total_bytes",
+        "total_compressed_bytes",
+    )
 
-    def __init__(self, protocol: BaseProtocol,
-                 *, limit: int=DEFAULT_LIMIT,
-                 timer: Optional[BaseTimerContext]=None,
-                 loop: Optional[asyncio.AbstractEventLoop]=None) -> None:
+    def __init__(
+        self,
+        protocol: BaseProtocol,
+        limit: int,
+        *,
+        timer: Optional[BaseTimerContext] = None,
+        loop: Optional[asyncio.AbstractEventLoop] = None,
+    ) -> None:
         self._protocol = protocol
         self._low_water = limit
         self._high_water = limit * 2
@@ -118,53 +149,63 @@ class StreamReader(AsyncStreamReaderMixin):
         self._loop = loop
         self._size = 0
         self._cursor = 0
-        self._http_chunk_splits = None  # type: Optional[List[int]]
-        self._buffer = collections.deque()  # type: Deque[bytes]
+        self._http_chunk_splits: Optional[List[int]] = None
+        self._buffer: Deque[bytes] = collections.deque()
         self._buffer_offset = 0
         self._eof = False
-        self._waiter = None  # type: Optional[asyncio.Future[None]]
-        self._eof_waiter = None  # type: Optional[asyncio.Future[None]]
-        self._exception = None  # type: Optional[BaseException]
-        self._timer = timer
-        self._eof_callbacks = []  # type: List[Callable[[], None]]
+        self._waiter: Optional[asyncio.Future[None]] = None
+        self._eof_waiter: Optional[asyncio.Future[None]] = None
+        self._exception: Optional[BaseException] = None
+        self._timer = TimerNoop() if timer is None else timer
+        self._eof_callbacks: List[Callable[[], None]] = []
+        self._eof_counter = 0
+        self.total_bytes = 0
+        self.total_compressed_bytes: Optional[int] = None
 
     def __repr__(self) -> str:
         info = [self.__class__.__name__]
         if self._size:
-            info.append('%d bytes' % self._size)
+            info.append("%d bytes" % self._size)
         if self._eof:
-            info.append('eof')
-        if self._low_water != DEFAULT_LIMIT:
-            info.append('low=%d high=%d' % (self._low_water, self._high_water))
+            info.append("eof")
+        if self._low_water != 2**16:  # default limit
+            info.append("low=%d high=%d" % (self._low_water, self._high_water))
         if self._waiter:
-            info.append('w=%r' % self._waiter)
+            info.append("w=%r" % self._waiter)
         if self._exception:
-            info.append('e=%r' % self._exception)
-        return '<%s>' % ' '.join(info)
+            info.append("e=%r" % self._exception)
+        return "<%s>" % " ".join(info)
+
+    def get_read_buffer_limits(self) -> Tuple[int, int]:
+        return (self._low_water, self._high_water)
 
     def exception(self) -> Optional[BaseException]:
         return self._exception
 
-    def set_exception(self, exc: BaseException) -> None:
+    def set_exception(
+        self,
+        exc: BaseException,
+        exc_cause: BaseException = _EXC_SENTINEL,
+    ) -> None:
         self._exception = exc
         self._eof_callbacks.clear()
 
         waiter = self._waiter
         if waiter is not None:
             self._waiter = None
-            set_exception(waiter, exc)
+            set_exception(waiter, exc, exc_cause)
 
         waiter = self._eof_waiter
         if waiter is not None:
             self._eof_waiter = None
-            set_exception(waiter, exc)
+            set_exception(waiter, exc, exc_cause)
 
     def on_eof(self, callback: Callable[[], None]) -> None:
         if self._eof:
             try:
                 callback()
             except Exception:
-                internal_logger.exception('Exception in eof callback')
+                internal_logger.exception("Exception in eof callback")
         else:
             self._eof_callbacks.append(callback)
 
@@ -181,11 +222,14 @@ class StreamReader(AsyncStreamReaderMixin):
             self._eof_waiter = None
             set_result(waiter, None)
 
+        if self._protocol._reading_paused:
+            self._protocol.resume_reading()
+
         for cb in self._eof_callbacks:
             try:
                 cb()
             except Exception:
-                internal_logger.exception('Exception in eof callback')
+                internal_logger.exception("Exception in eof callback")
 
         self._eof_callbacks.clear()
 
@@ -208,18 +252,25 @@ class StreamReader(AsyncStreamReaderMixin):
         finally:
             self._eof_waiter = None
 
+    @property
+    def total_raw_bytes(self) -> int:
+        if self.total_compressed_bytes is None:
+            return self.total_bytes
+        return self.total_compressed_bytes
+
     def unread_data(self, data: bytes) -> None:
-        """ rollback reading some data from stream, inserting it to buffer head.
-        """
-        warnings.warn("unread_data() is deprecated "
-                      "and will be removed in future releases (#3260)",
-                      DeprecationWarning,
-                      stacklevel=2)
+        """rollback reading some data from stream, inserting it to buffer head."""
+        warnings.warn(
+            "unread_data() is deprecated "
+            "and will be removed in future releases (#3260)",
+            DeprecationWarning,
+            stacklevel=2,
+        )
         if not data:
             return
 
         if self._buffer_offset:
-            self._buffer[0] = self._buffer[0][self._buffer_offset:]
+            self._buffer[0] = self._buffer[0][self._buffer_offset :]
             self._buffer_offset = 0
         self._size += len(data)
         self._cursor -= len(data)
@@ -227,41 +278,44 @@ class StreamReader(AsyncStreamReaderMixin):
         self._eof_counter = 0
 
     # TODO: size is ignored, remove the param later
-    def feed_data(self, data: bytes, size: int=0) -> None:
-        assert not self._eof, 'feed_data after feed_eof'
+    def feed_data(self, data: bytes, size: int = 0) -> None:
+        assert not self._eof, "feed_data after feed_eof"
 
         if not data:
             return
 
-        self._size += len(data)
+        data_len = len(data)
+        self._size += data_len
         self._buffer.append(data)
-        self.total_bytes += len(data)
+        self.total_bytes += data_len
 
         waiter = self._waiter
         if waiter is not None:
             self._waiter = None
             set_result(waiter, None)
 
-        if (self._size > self._high_water and
-                not self._protocol._reading_paused):
+        if self._size > self._high_water and not self._protocol._reading_paused:
             self._protocol.pause_reading()
 
     def begin_http_chunk_receiving(self) -> None:
         if self._http_chunk_splits is None:
             if self.total_bytes:
-                raise RuntimeError("Called begin_http_chunk_receiving when"
-                                   "some data was already fed")
+                raise RuntimeError(
+                    "Called begin_http_chunk_receiving when some data was already fed"
+                )
             self._http_chunk_splits = []
 
     def end_http_chunk_receiving(self) -> None:
         if self._http_chunk_splits is None:
-            raise RuntimeError("Called end_chunk_receiving without calling "
-                               "begin_chunk_receiving first")
+            raise RuntimeError(
+                "Called end_chunk_receiving without calling "
+                "begin_chunk_receiving first"
+            )
 
         # self._http_chunk_splits contains logical byte offsets from start of
         # the body transfer. Each offset is the offset of the end of a chunk.
         # "Logical" means bytes, accessible for a user.
-        # If no chunks containig logical data were received, current position
+        # If no chunks containing logical data were received, current position
         # is difinitely zero.
         pos = self._http_chunk_splits[-1] if self._http_chunk_splits else 0
 
@@ -282,55 +336,66 @@ class StreamReader(AsyncStreamReaderMixin):
             set_result(waiter, None)
 
     async def _wait(self, func_name: str) -> None:
+        if not self._protocol.connected:
+            raise RuntimeError("Connection closed.")
+
         # StreamReader uses a future to link the protocol feed_data() method
         # to a read coroutine. Running two read coroutines at the same time
         # would have an unexpected behaviour. It would not possible to know
         # which coroutine would get the next data.
         if self._waiter is not None:
-            raise RuntimeError('%s() called while another coroutine is '
-                               'already waiting for incoming data' % func_name)
+            raise RuntimeError(
+                "%s() called while another coroutine is "
+                "already waiting for incoming data" % func_name
+            )
 
         waiter = self._waiter = self._loop.create_future()
         try:
-            if self._timer:
-                with self._timer:
-                    await waiter
-            else:
+            with self._timer:
                 await waiter
         finally:
             self._waiter = None
 
     async def readline(self) -> bytes:
+        return await self.readuntil()
+
+    async def readuntil(self, separator: bytes = b"\n") -> bytes:
+        seplen = len(separator)
+        if seplen == 0:
+            raise ValueError("Separator should be at least one-byte string")
+
         if self._exception is not None:
             raise self._exception
 
-        line = []
-        line_size = 0
+        chunk = b""
+        chunk_size = 0
         not_enough = True
 
         while not_enough:
             while self._buffer and not_enough:
                 offset = self._buffer_offset
-                ichar = self._buffer[0].find(b'\n', offset) + 1
-                # Read from current offset to found b'\n' or to the end.
-                data = self._read_nowait_chunk(ichar - offset if ichar else -1)
-                line.append(data)
-                line_size += len(data)
+                ichar = self._buffer[0].find(separator, offset) + 1
+                # Read from current offset to found separator or to the end.
+                data = self._read_nowait_chunk(
+                    ichar - offset + seplen - 1 if ichar else -1
+                )
+                chunk += data
+                chunk_size += len(data)
                 if ichar:
                     not_enough = False
 
-                if line_size > self._high_water:
-                    raise ValueError('Line is too long')
+                if chunk_size > self._high_water:
+                    raise ValueError("Chunk too big")
 
             if self._eof:
                 break
 
             if not_enough:
-                await self._wait('readline')
+                await self._wait("readuntil")
 
-        return b''.join(line)
+        return chunk
 
-    async def read(self, n: int=-1) -> bytes:
+    async def read(self, n: int = -1) -> bytes:
         if self._exception is not None:
             raise self._exception
 
@@ -340,14 +405,16 @@ class StreamReader(AsyncStreamReaderMixin):
         # lets keep this code one major release.
         if __debug__:
             if self._eof and not self._buffer:
-                self._eof_counter = getattr(self, '_eof_counter', 0) + 1
+                self._eof_counter = getattr(self, "_eof_counter", 0) + 1
                 if self._eof_counter > 5:
                     internal_logger.warning(
-                        'Multiple access to StreamReader in eof state, '
-                        'might be infinite loop.', stack_info=True)
+                        "Multiple access to StreamReader in eof state, "
+                        "might be infinite loop.",
+                        stack_info=True,
+                    )
 
         if not n:
-            return b''
+            return b""
 
         if n < 0:
             # This used to just loop creating a new waiter hoping to
@@ -360,13 +427,13 @@ class StreamReader(AsyncStreamReaderMixin):
                 if not block:
                     break
                 blocks.append(block)
-            return b''.join(blocks)
+            return b"".join(blocks)
 
         # TODO: should be `if` instead of `while`
         # because waiter maybe triggered on chunk end,
         # without feeding any data
         while not self._buffer and not self._eof:
-            await self._wait('read')
+            await self._wait("read")
 
         return self._read_nowait(n)
 
@@ -378,12 +445,14 @@ class StreamReader(AsyncStreamReaderMixin):
         # because waiter maybe triggered on chunk end,
         # without feeding any data
         while not self._buffer and not self._eof:
-            await self._wait('readany')
+            await self._wait("readany")
 
         return self._read_nowait(-1)
 
     async def readchunk(self) -> Tuple[bytes, bool]:
-        """Returns a tuple of (data, end_of_http_chunk). When chunked transfer
+        """Returns a tuple of (data, end_of_http_chunk).
+
+        When chunked transfer
         encoding is used, end_of_http_chunk is a boolean indicating if the end
         of the data corresponds to the end of a HTTP chunk , otherwise it is
         always False.
@@ -397,9 +466,11 @@ class StreamReader(AsyncStreamReaderMixin):
                 if pos == self._cursor:
                     return (b"", True)
                 if pos > self._cursor:
-                    return (self._read_nowait(pos-self._cursor), True)
-                internal_logger.warning('Skipping HTTP chunk end due to data '
-                                        'consumption beyond chunk boundary')
+                    return (self._read_nowait(pos - self._cursor), True)
+                internal_logger.warning(
+                    "Skipping HTTP chunk end due to data "
+                    "consumption beyond chunk boundary"
+                )
 
             if self._buffer:
                 return (self._read_nowait_chunk(-1), False)
@@ -408,27 +479,26 @@ class StreamReader(AsyncStreamReaderMixin):
             if self._eof:
                 # Special case for signifying EOF.
                 # (b'', True) is not a final return value actually.
-                return (b'', False)
+                return (b"", False)
 
-            await self._wait('readchunk')
+            await self._wait("readchunk")
 
     async def readexactly(self, n: int) -> bytes:
         if self._exception is not None:
             raise self._exception
 
-        blocks = []  # type: List[bytes]
+        blocks: List[bytes] = []
         while n > 0:
             block = await self.read(n)
             if not block:
-                partial = b''.join(blocks)
-                raise asyncio.streams.IncompleteReadError(
-                    partial, len(partial) + n)
+                partial = b"".join(blocks)
+                raise asyncio.IncompleteReadError(partial, len(partial) + n)
             blocks.append(block)
             n -= len(block)
 
-        return b''.join(blocks)
+        return b"".join(blocks)
 
-    def read_nowait(self, n: int=-1) -> bytes:
+    def read_nowait(self, n: int = -1) -> bytes:
         # default was changed to be consistent with .read(-1)
         #
         # I believe the most users don't know about the method and
@@ -438,7 +508,8 @@ class StreamReader(AsyncStreamReaderMixin):
 
         if self._waiter and not self._waiter.done():
             raise RuntimeError(
-                'Called while some coroutine is waiting for incoming data.')
+                "Called while some coroutine is waiting for incoming data."
+            )
 
         return self._read_nowait(n)
 
@@ -446,7 +517,7 @@ class StreamReader(AsyncStreamReaderMixin):
         first_buffer = self._buffer[0]
         offset = self._buffer_offset
         if n != -1 and len(first_buffer) - offset > n:
-            data = first_buffer[offset:offset + n]
+            data = first_buffer[offset : offset + n]
             self._buffer_offset += n
 
         elif offset:
@@ -457,8 +528,9 @@ class StreamReader(AsyncStreamReaderMixin):
         else:
             data = self._buffer.popleft()
 
-        self._size -= len(data)
-        self._cursor += len(data)
+        data_len = len(data)
+        self._size -= data_len
+        self._cursor += data_len
 
         chunk_splits = self._http_chunk_splits
         # Prevent memory leak: drop useless chunk splits
@@ -470,9 +542,10 @@ class StreamReader(AsyncStreamReaderMixin):
         return data
 
     def _read_nowait(self, n: int) -> bytes:
-        """ Read not more than n bytes, or whole buffer is n == -1 """
-        chunks = []
+        """Read not more than n bytes, or whole buffer if n == -1"""
+        self._timer.assert_timeout()
 
+        chunks = []
         while self._buffer:
             chunk = self._read_nowait_chunk(n)
             chunks.append(chunk)
@@ -481,22 +554,35 @@ class StreamReader(AsyncStreamReaderMixin):
                 if n == 0:
                     break
 
-        return b''.join(chunks) if chunks else b''
+        return b"".join(chunks) if chunks else b""
 
 
-class EmptyStreamReader(AsyncStreamReaderMixin):
+class EmptyStreamReader(StreamReader):  # lgtm [py/missing-call-to-init]
+
+    __slots__ = ("_read_eof_chunk",)
+
+    def __init__(self) -> None:
+        self._read_eof_chunk = False
+        self.total_bytes = 0
+
+    def __repr__(self) -> str:
+        return "<%s>" % self.__class__.__name__
 
     def exception(self) -> Optional[BaseException]:
         return None
 
-    def set_exception(self, exc: BaseException) -> None:
+    def set_exception(
+        self,
+        exc: BaseException,
+        exc_cause: BaseException = _EXC_SENTINEL,
+    ) -> None:
         pass
 
     def on_eof(self, callback: Callable[[], None]) -> None:
         try:
             callback()
         except Exception:
-            internal_logger.exception('Exception in eof callback')
+            internal_logger.exception("Exception in eof callback")
 
     def feed_eof(self) -> None:
         pass
@@ -510,29 +596,35 @@ class EmptyStreamReader(AsyncStreamReaderMixin):
     async def wait_eof(self) -> None:
         return
 
-    def feed_data(self, data: bytes, n: int=0) -> None:
+    def feed_data(self, data: bytes, n: int = 0) -> None:
         pass
 
     async def readline(self) -> bytes:
-        return b''
+        return b""
 
-    async def read(self, n: int=-1) -> bytes:
-        return b''
+    async def read(self, n: int = -1) -> bytes:
+        return b""
+
+    # TODO add async def readuntil
 
     async def readany(self) -> bytes:
-        return b''
+        return b""
 
     async def readchunk(self) -> Tuple[bytes, bool]:
-        return (b'', True)
+        if not self._read_eof_chunk:
+            self._read_eof_chunk = True
+            return (b"", False)
+
+        return (b"", True)
 
     async def readexactly(self, n: int) -> bytes:
-        raise asyncio.streams.IncompleteReadError(b'', n)
+        raise asyncio.IncompleteReadError(b"", n)
 
-    def read_nowait(self) -> bytes:
-        return b''
+    def read_nowait(self, n: int = -1) -> bytes:
+        return b""
 
 
-EMPTY_PAYLOAD = EmptyStreamReader()
+EMPTY_PAYLOAD: Final[StreamReader] = EmptyStreamReader()
 
 
 class DataQueue(Generic[_T]):
@@ -541,10 +633,9 @@ class DataQueue(Generic[_T]):
     def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
         self._loop = loop
         self._eof = False
-        self._waiter = None  # type: Optional[asyncio.Future[None]]
-        self._exception = None  # type: Optional[BaseException]
-        self._size = 0
-        self._buffer = collections.deque()  # type: Deque[Tuple[_T, int]]
+        self._waiter: Optional[asyncio.Future[None]] = None
+        self._exception: Optional[BaseException] = None
+        self._buffer: Deque[Tuple[_T, int]] = collections.deque()
 
     def __len__(self) -> int:
         return len(self._buffer)
@@ -558,29 +649,26 @@ class DataQueue(Generic[_T]):
     def exception(self) -> Optional[BaseException]:
         return self._exception
 
-    def set_exception(self, exc: BaseException) -> None:
+    def set_exception(
+        self,
+        exc: BaseException,
+        exc_cause: BaseException = _EXC_SENTINEL,
+    ) -> None:
         self._eof = True
         self._exception = exc
-
-        waiter = self._waiter
-        if waiter is not None:
+        if (waiter := self._waiter) is not None:
             self._waiter = None
-            set_exception(waiter, exc)
+            set_exception(waiter, exc, exc_cause)
 
-    def feed_data(self, data: _T, size: int=0) -> None:
-        self._size += size
+    def feed_data(self, data: _T, size: int = 0) -> None:
         self._buffer.append((data, size))
-
-        waiter = self._waiter
-        if waiter is not None:
+        if (waiter := self._waiter) is not None:
             self._waiter = None
             set_result(waiter, None)
 
     def feed_eof(self) -> None:
         self._eof = True
-
-        waiter = self._waiter
-        if waiter is not None:
+        if (waiter := self._waiter) is not None:
             self._waiter = None
             set_result(waiter, None)
 
@@ -593,16 +681,12 @@ class DataQueue(Generic[_T]):
             except (asyncio.CancelledError, asyncio.TimeoutError):
                 self._waiter = None
                 raise
-
         if self._buffer:
-            data, size = self._buffer.popleft()
-            self._size -= size
+            data, _ = self._buffer.popleft()
             return data
-        else:
-            if self._exception is not None:
-                raise self._exception
-            else:
-                raise EofStream
+        if self._exception is not None:
+            raise self._exception
+        raise EofStream
 
     def __aiter__(self) -> AsyncStreamIterator[_T]:
         return AsyncStreamIterator(self.read)
@@ -611,25 +695,41 @@ class DataQueue(Generic[_T]):
 class FlowControlDataQueue(DataQueue[_T]):
     """FlowControlDataQueue resumes and pauses an underlying stream.
 
-    It is a destination for parsed data."""
+    It is a destination for parsed data.
 
-    def __init__(self, protocol: BaseProtocol, *,
-                 limit: int=DEFAULT_LIMIT,
-                 loop: asyncio.AbstractEventLoop) -> None:
+    This class is deprecated and will be removed in version 4.0.
+    """
+
+    def __init__(
+        self, protocol: BaseProtocol, limit: int, *, loop: asyncio.AbstractEventLoop
+    ) -> None:
         super().__init__(loop=loop)
-
+        self._size = 0
         self._protocol = protocol
         self._limit = limit * 2
 
-    def feed_data(self, data: _T, size: int=0) -> None:
+    def feed_data(self, data: _T, size: int = 0) -> None:
         super().feed_data(data, size)
+        self._size += size
 
         if self._size > self._limit and not self._protocol._reading_paused:
             self._protocol.pause_reading()
 
     async def read(self) -> _T:
-        try:
-            return await super().read()
-        finally:
+        if not self._buffer and not self._eof:
+            assert not self._waiter
+            self._waiter = self._loop.create_future()
+            try:
+                await self._waiter
+            except (asyncio.CancelledError, asyncio.TimeoutError):
+                self._waiter = None
+                raise
+        if self._buffer:
+            data, size = self._buffer.popleft()
+            self._size -= size
             if self._size < self._limit and self._protocol._reading_paused:
                 self._protocol.resume_reading()
+            return data
+        if self._exception is not None:
+            raise self._exception
+        raise EofStream
